#!/usr/bin/env python3

# Description: TCP/IP server to display the Image for Windows/Linux operation 
#              status when using the "--msvr" option.
#
# Version: 1.00
#
# Requires: Image for Windows/Linux 3.53 or later
#           Python 3.9 or later
#
# Optional: ncurses
#
# Copyright (c) 2022 TeraByte, Inc.  All Rights Reserved.
#
# Disclaimer:
#
# This script is provided AS IS without warranty of any kind. TeraByte, Inc.
# further disclaims all implied warranties including, without limitation,
# any implied warranties of merchantability or of fitness for a particular
# purpose. The entire risk arising out of the use or performance of the
# sample scripts and documentation remains with you. In no event shall
# TeraByte, Inc., its authors, or anyone else involved in the creation,
# production, or delivery of the scripts be liable for any damages whatsoever
# (including, without limitation, damages for loss of business profits,
# business interruption, loss of business information, or other pecuniary
# loss) arising out of the use of or inability to use the sample scripts or
# documentation, even if TeraByte, Inc. has been advised of the possibility
# of such damages.
#
# Permission to use, copy, modify, and distribute this script for use with
# TeraByte products with or without fee is hereby granted, provided that the 
# above copyright notice, disclaimer and this permission notice appear in all 
# copies, and that the name of TeraByte Unlimited or TeraByte, Inc. not be 
# used in advertising or publicity pertaining to distribution of the script 
# without specific, written prior permission.
#

import sys
import os
import concurrent.futures
import socket
import queue
import threading
import select
import time
import argparse
if sys.platform == "win32":
    import msvcrt
try:
    import curses
    use_curses = True
except ImportError:
    use_curses = False


# vars
qMsg = queue.Queue()
EventExit = threading.Event()
EventConnected = []
EventCompleted = []
EventCancel = []

class ClientData:
    def __init__(self):
        self.Reset()
    def Reset(self):
        self.info = []
        self.stats = []
        self.who = "Image for Windows/Linux"
        self.ip = ""

class ResultData:
    def __init__(self, timestr, result, code, ip):
        self.timestr = timestr
        self.result = result
        self.code = code
        self.ip = ip

def ScreenThread():
    """ update screen and process input """
    
    cdata = []
    cdata.append(ClientData())
    results = []

    update = True
    msg = []
    autoupdate = 0
    dispmode = 0  # display mode: 0=client, 1=summary, 2=results
    curindex = 0  # current index
    msgindex = 0  # message index
    UpdateSummary = False
    UpdateSummaryRR = gRefreshRate
    if UpdateSummaryRR > 10:
        UpdateSummaryRR = 10
    ResultsFn = os.path.join(os.path.abspath(os.path.dirname(__file__)),"tbmsvr_results.log")

    while not EventExit.is_set() or not qMsg.empty():
        
        ClientCount = len(cdata)

        # check for input
        kb = GetKeyCommand()
        if kb == "F2":      # exit script
            EventExit.set()
            update = True
        elif kb == "ENTER": # client/summary screen toggle
            if dispmode == 0 or dispmode == 2:
                dispmode = 1
            else:
                dispmode = 0
            update = True
        elif kb == "F4":    # results screen
            if dispmode == 1:
                dispmode = 2
                update = True
            elif dispmode == 2:
                dispmode = 1
                update = True
        elif kb == "F5":    # refresh screen
            update = True
        elif kb == "F12" and dispmode == 0:   # cancel client operation
            if EventConnected[curindex].is_set():
                EventCancel[curindex].set()
                update = True
        elif kb == "CTRL-F12" and (dispmode == 0 or dispmode == 1):  # exit client thread
            if EventConnected[curindex].is_set():
                EventCompleted[curindex].set()
        elif kb == "LEFT" or kb == "UP":  # left/up arrow (previous client)
            curindex -= 1
            if curindex < 0:
                curindex = 0
            update = True
        elif kb == "RIGHT" or kb == "DOWN": # right/down arrow (next client)
            curindex += 1
            if curindex > ClientCount-1:
                curindex = ClientCount-1
            update = True

        # check for messages
        msgcnt = 0
        while not qMsg.empty():
            msgcnt += 1
            if msgcnt > 10:
                break
            msg = qMsg.get()
            msgindex = int(msg[0])
            while msgindex > ClientCount-1:
                cdata.append(ClientData())
                ClientCount = len(cdata)
            if msgindex < 0:
                continue
            if msg[1] == "WHO:":
                cdata[msgindex].who = msg[2]
                cdata[msgindex].ip = msg[3]
                if dispmode == 0 and msgindex == curindex:
                    update = True
            elif msg[1] == "INFO:":
                cdata[msgindex].info = msg[2:]
                if (dispmode == 0 and msgindex == curindex) or dispmode == 1:
                    update = True
            elif msg[1] == "STATS:":
                cdata[msgindex].stats = msg[2:]
                UpdateSummary = True
                if dispmode == 0 and msgindex == curindex:
                    update = True
            elif msg[1] == "COMPLETED:":
                restime = time.localtime()
                resdata = ResultData(time.strftime('%a %b %d %H:%M',restime),msg[2],int(msg[3]),cdata[msgindex].ip)
                results.append(resdata)
                if len(results) > 15:
                    del(results[0])
                update = True
                # save result
                try:
                    with open(ResultsFn, 'a', encoding='utf-8') as f:
                        f.write(f"{time.strftime('%Y-%m-%d %H:%M:%S',restime)} - {resdata.ip:15} - {resdata.result}\n")
                except:
                    pass
            elif msg[1] == "RESET:":
                cdata[msgindex].Reset()
                if (dispmode == 0 and msgindex == curindex) or dispmode == 1:
                    update = True

        # update screen
        if use_curses and autoupdate > 40:
            update = True
        elif dispmode == 1 and autoupdate >= UpdateSummaryRR*10 and UpdateSummary == True:
            update = True
            UpdateSummary = False
        if not update:
            time.sleep(0.1)
            autoupdate += 1
        else:
            update = False
            autoupdate = 0
            WaitStr = "Waiting..."

            # curses used (Windows or Linux)
            if use_curses:
                curses.curs_set(0)
                stdscr.erase()
                # results ----------------------------------------------------------------------------
                if dispmode == 2:
                    cursesprint(0,1,f" {'Image for Windows/Linux Operation Status':56}    Client Results ", 1)
                    y = 3
                    if len(results) == 0:
                        cursesprint(0,y,"No results")
                        y += 1
                    else:
                        for i in range(len(results)):
                            s = f"{results[i].ip} - {results[i].timestr} - "
                            cursesprint(0,y,s)
                            if results[i].code == 0: # success
                                rescol = 2
                            elif results[i].code == 1:   # canceled
                                rescol = 3
                            else:   # error
                                rescol = 4
                            cursesprint(0+len(s),y,f"{results[i].result}", rescol, curses.A_BOLD)
                            y += 1
                    y += 1
                    cursesprint(0,y,f"{' F2 - Exit | Enter - Summary':76}", 1)
                    stdscr.refresh()
                # summary ----------------------------------------------------------------------------
                elif dispmode == 1:
                    cursesprint(0,1,f" {'Image for Windows/Linux Operation Status':56}    Client Summary ", 1)
                    y = 3
                    for i in range(ClientCount):
                        selch = ">" if i == curindex else " "
                        info = WaitStr if len(cdata[i].info) == 0 else cdata[i].info[0]
                        if EventConnected[i].is_set():
                            prg = "  0%" if len(cdata[i].stats) < 5 else f"{int(cdata[i].stats[1]):3d}%"
                            cursesprint(0,y,f"{selch} #{i+1:>2} - ")
                            cursesprint(8,y,f"{prg}", 2, curses.A_BOLD)
                            cursesprint(8+len(prg),y,f" {info}")
                        else:
                            cursesprint(0,y,f"{selch} #{i+1:>2} - {info}")
                        y += 1
                    y += 1
                    cursesprint(0,y,f" \u2191\u2193 {'- Select Client | Enter - Client View | F2 - Exit | F4 - Results':72}", 1)
                    stdscr.refresh()
                # client data ----------------------------------------------------------------------------
                else:
                    cursesprint(0,1,f" {'Image for Windows/Linux Operation Status':53}      Client {curindex+1:>2} of {ClientCount:>2} ", 1)
                    y = 3
                    if EventConnected[curindex].is_set():
                        cursesprint(0,y,f"PROGRAM: {cdata[curindex].who}")
                        cursesprint(0,y+1,"-")
                        y += 2
                    if EventCancel[curindex].is_set():
                        s = ">>>>> Operation Canceled <<<<<"
                        if not EventCompleted[curindex].is_set():
                            s = s + "    Please wait..."
                        cursesprint(0,y,s,3,curses.A_BOLD)
                        cursesprint(0,y+1,"-")
                        y += 2
                    if len(cdata[curindex].info) == 0:
                        cursesprint(0,y,WaitStr)
                        y += 2
                    else:
                        for s in cdata[curindex].info:
                            cursesprint(0,y,s)
                            y += 1
                    if len(cdata[curindex].stats) >= 5:
                        posProgress = int(cdata[curindex].stats[0])
                        posProgressBar = int(posProgress/2)
                        posOverall = int(cdata[curindex].stats[1])
                        posOverallBar = int(posOverall/2)
                        cursesprint(0,y,"-")
                        cursesprint(0,y+1,f"   Item Progress: {posProgress:3d}% ")
                        cursesprint(23,y+1,f"[{'#'*(posProgressBar)}{' '*(50-posProgressBar)}]", 2)
                        cursesprint(0,y+2,f"Overall Progress: {posOverall:3d}% ")
                        cursesprint(23,y+2,f"[{'#'*(posOverallBar)}{' '*(50-posOverallBar)}]", 2)
                        cursesprint(0,y+3,f" MiB Transferred: {cdata[curindex].stats[2]}")
                        cursesprint(0,y+4,f"   MiB Remaining: {cdata[curindex].stats[3]}")
                        cursesprint(0,y+5,f"    Time Elapsed: {cdata[curindex].stats[4]}")
                        cursesprint(0,y+6,f"  Time Remaining: {cdata[curindex].stats[5]}")
                        cursesprint(0,y+8,f" \u2191\u2193 {'- Select Client | Enter - Summary | F2 - Exit | F12 - Cancel Operation':72}", 1)
                    else:
                        for s in cdata[curindex].stats:
                            y += 1
                            cursesprint(0,y,s)
                        if len(cdata[curindex].stats) > 0:
                            y += 1
                        if not EventConnected[curindex].is_set():
                            cursesprint(0,y,f" \u2191\u2193 {'- Select Client | Enter - Summary | F2 - Exit':72}", 1)            
                    stdscr.refresh()

            # curses NOT used (Windows)                                   
            else:
                os.system("cls")
                # results ----------------------------------------------------------------------------
                if dispmode == 2:
                    print("="*76)
                    print(f"  {'Image for Windows/Linux Operation Status':53}    Client Results")
                    print("="*76)
                    print("")
                    if len(results) == 0:
                        print("No results")
                    else:
                        for i in range(len(results)):
                            print(results[i].ip, "-", results[i].timestr, "-", results[i].result)
                    print("")
                    print("-"*76)
                    print("F2 - Exit | Enter - Summary")
                # summary ----------------------------------------------------------------------------
                elif dispmode == 1:
                    print("="*76)
                    print(f"  {'Image for Windows/Linux Operation Status':53}    Client Summary")
                    print("="*76)
                    print("")
                    for i in range(ClientCount):
                        selch = ">" if i == curindex else " "
                        info = WaitStr if len(cdata[i].info) == 0 else cdata[i].info[0]
                        if EventConnected[i].is_set():
                            prg = "  0%" if len(cdata[i].stats) < 5 else f"{int(cdata[i].stats[1]):3d}%"
                            print(f"{selch} #{i+1:>2} - {prg} {info}")
                        else:
                            print(f"{selch} #{i+1:>2} - {info}")
                    print("")
                    print("-"*76)
                    print("\u2191\u2193 - Select Client | Enter - Client View | F2 - Exit | F4 - Results")
                # client data ----------------------------------------------------------------------------
                else:
                    print("="*76)
                    print(f"  {'Image for Windows/Linux Operation Status':53}    Client {curindex+1} of {ClientCount}")
                    print("="*76)
                    print("")
                    if EventConnected[curindex].is_set():
                        print(f"PROGRAM: {cdata[curindex].who}")
                        print("-")
                    if EventCancel[curindex].is_set():
                        s = ">>>>> Operation Canceled <<<<<"
                        if not EventCompleted[curindex].is_set():
                            s = s + "    Please wait..."
                        print(s)
                        print("-")
                    if len(cdata[curindex].info) == 0:
                        print(WaitStr)
                        print("")
                    else:
                        for s in cdata[curindex].info:
                            print(s)
                    if len(cdata[curindex].stats) >= 5:
                        posProgress = int(cdata[curindex].stats[0])
                        posProgressBar = int(posProgress/2)
                        posOverall = int(cdata[curindex].stats[1])
                        posOverallBar = int(posOverall/2)
                        print("-")
                        print(f"   Item Progress: {posProgress:3d}% [{'#'*(posProgressBar)}{' '*(50-posProgressBar)}]", sep='')
                        print(f"Overall Progress: {posOverall:3d}% [{'#'*(posOverallBar)}{' '*(50-posOverallBar)}]", sep='')
                        print(" MiB Transferred:", cdata[curindex].stats[2])
                        print("   MiB Remaining:", cdata[curindex].stats[3])
                        print("    Time Elapsed:", cdata[curindex].stats[4])
                        print("  Time Remaining:", cdata[curindex].stats[5])
                        print("")
                        print("-"*76)
                        print("\u2191\u2193 - Select Client | Enter - Summary | F2 - Exit | F12 - Cancel Operation")
                    else:
                        if len(cdata[curindex].stats) > 0:
                            print("")
                            for s in cdata[curindex].stats:
                                print(s)
                            print("")
                        if not EventConnected[curindex].is_set():
                            print("-"*76)
                            print("\u2191\u2193 - Select Client | Enter - Summary | F2 - Exit")

    # screen thread end

   
class ClientThread(threading.Thread):

    def __init__(self, conn:socket.socket, ip, port, evindex):
        threading.Thread.__init__(self)
        self.conn = conn
        self.ip = ip
        self.port = port
        self.evindex = evindex
        
        self.conn.setblocking(True)
        self.rfile = self.conn.makefile('rb',-1)

    def run(self):
        msg = []
        CancelSent = False
        InfoReceived = False
        WhoReceived = False
        StatsReceived = False
        evindex = self.evindex
        EventConnected[evindex].set()

        try:
            qMsg.put([f"{evindex}","INFO:","Connected"])
            s = f"R_WHO\r\nR_INFO\r\nR_AUTO {gRefreshRate}\r\n"
            self.conn.sendall(s.encode('utf-8'))

            while not EventCompleted[evindex].is_set() and not EventExit.is_set():
                # check for cancel
                if not CancelSent and EventCancel[evindex].is_set():
                    try:
                        self.conn.sendall(b"A_CANCEL\r\n")
                        CancelSent = True
                    except:
                        break   # exit on error

                # read message
                rlist,wlist,elist = select.select([self.rfile],[],[],0.5)
                if self.rfile in rlist:
                    try:
                        self.data = self.rfile.readline()
                    except ConnectionError:
                        break   # exit on error
                else:
                    if not InfoReceived:
                        self.conn.sendall(b"R_INFO\r\n")
                    if not WhoReceived:
                        self.conn.sendall(b"R_WHO\r\n")
                    if not StatsReceived:
                        s = f"R_AUTO {gRefreshRate}\r\n"
                        self.conn.sendall(s.encode('utf-8'))
                    continue

                # process message
                if len(self.data) == 0 or self.conn.fileno() == -1:
                    break   # exit on no data/error
                cmd = str(self.data, 'utf-8').strip()
                if cmd[:6] == "STATS:":
                    cnt = int(cmd[7:])
                    msg.clear()
                    msg.append(f"{evindex}")
                    msg.append("STATS:")
                    for i in range(cnt):
                        self.data = self.rfile.readline().strip()
                        s = str(self.data, 'utf-8')
                        msg.append(s)
                    qMsg.put(list(msg))
                    StatsReceived = True
                elif cmd[:5] == "INFO:":
                    cnt = int(cmd[6:])
                    msg.clear()
                    msg.append(f"{evindex}")
                    msg.append("INFO:")
                    for i in range(cnt):
                        self.data = self.rfile.readline().strip()
                        s = str(self.data, 'utf-8')
                        msg.append(s)
                    qMsg.put(list(msg))
                    InfoReceived = True
                elif cmd[:4] == "WHO:":
                    self.data = self.rfile.readline().strip()
                    s = str(self.data, 'utf-8')
                    qMsg.put([f"{evindex}","WHO:",f"{s}  -  ({self.ip}, Port: {self.port})",f"{self.ip}"])
                    WhoReceived = True
                elif cmd[:10] == "COMPLETED:":
                    res = int(cmd[11:])
                    if res == 0:
                        qMsg.put([f"{evindex}","COMPLETED:",f"{cmd} >> Success <<",f"{res}"])
                    elif res == 1:
                        qMsg.put([f"{evindex}","COMPLETED:",f"{cmd} >> Operation Canceled <<",f"{res}"])
                    else:
                        qMsg.put([f"{evindex}","COMPLETED:",f"{cmd} >> Error <<",f"{res}"])
                    EventCompleted[evindex].set()
        except:
            raise
        finally:
            # cleanup
            self.rfile.close()
            self.conn.close()
            EventCompleted[evindex].clear()
            EventCancel[evindex].clear()
            EventConnected[evindex].clear()
            qMsg.put([f"{evindex}","RESET:"])


def ServerThread():
    """ server thread """

    # get host name, IP address, etc.
    host = socket.gethostname()
    hostIP = socket.gethostbyname(host)
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        s.connect(('10.255.255.255', 1))
        hostIP = s.getsockname()[0]
    except Exception:
        hostIP = ''
    finally:
        s.close()

    # create server
    server = socket.socket()
    server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    server.bind((hostIP,gPort))
    server.settimeout(1.0)
    server.listen(4)

    # connect and launch client threads
    clthreads = []
    evindex = 0
    SendWaitMsg = True
    while not EventExit.is_set():
        oldevindex = evindex
        evindex = len(clthreads)
        for i in range(len(clthreads)):
            if not clthreads[i].is_alive():
                evindex = i
                break
        if evindex+1 > len(EventConnected):
            EventConnected.append(threading.Event())
            EventCompleted.append(threading.Event())
            EventCancel.append(threading.Event())
        if evindex != oldevindex:
            if not EventConnected[oldevindex].is_set():
                qMsg.put([f"{oldevindex}","RESET:"])
            SendWaitMsg = True
        if SendWaitMsg:
            qMsg.put([f"{evindex}","INFO:",f"Waiting for connection...  -  [{host}:{hostIP}, {gPort}]",""])
            SendWaitMsg = False
        try:
            (conn, (ip,port)) = server.accept()
        except socket.timeout:
            continue
        client = ClientThread(conn,ip,port,evindex)
        client.start()
        if evindex+1 > len(clthreads):
            clthreads.append(client)
        else:
            clthreads[evindex] = client


def cursesprint(x:int, y:int, s:str, cp:int=0, attr:int=0):
    """ print string to curses window, ignore error """
    try:
        stdscr.addstr(y,x,s,curses.color_pair(cp) | attr)
    except curses.error:
        pass


def GetKeyCommand():
    """ Return specific key presses """

    if use_curses:
        stdscr.nodelay(True)
        ch = stdscr.getch()
        if ch == curses.KEY_F2:     # F2 - Exit
            return "F2"
        elif ch == 10:              # Enter - Summary toggle
            return "ENTER"
        elif ch == curses.KEY_F4:   # F4 - Results
            return "F4"
        elif ch == curses.KEY_F5:   # F5 - Refresh screen
            return "F5"
        elif ch == curses.KEY_F12:  # F12 - Cancel operation
            return "F12"
        elif ch == 300:             # Ctrl-F12 - exit client thread
            return "CTRL-F12"
        elif ch == curses.KEY_UP:   # Up-arrow
            return "UP"
        elif ch == curses.KEY_DOWN: # Down-arrow
            return "DOWN"
        elif ch == curses.KEY_LEFT: # Left-arrow
            return "LEFT"
        elif ch == curses.KEY_RIGHT:    # Right-arrow
            return "RIGHT"
    else:
        if msvcrt.kbhit():
            ch = msvcrt.getch()
            if ord(ch) == 13:       # Enter
                return "ENTER"
            elif ord(ch) == 0:
                ch = msvcrt.getch()
                if ord(ch) == 60:   # F2
                    return "F2"
                elif ord(ch) == 62: # F4
                    return "F4"
                elif ord(ch) == 63: # F5
                    return "F5"
            elif ord(ch) == 224:
                ch = msvcrt.getch()
                if ord(ch) == 134:   # F12
                    return "F12"
                elif ord(ch) == 138: # Ctrl-F12
                    return "CTRL-F12"
                elif ord(ch) == 72:  # Up-arrow
                    return "UP"
                elif ord(ch) == 75:  # Left-arrow
                    return "LEFT"
                elif ord(ch) == 77:  # Right-arrow
                    return "RIGHT"
                elif ord(ch) == 80:  # Down-arrow
                    return "DOWN"

    return ""   # no key to return


# =========================================================================================
# main
def main(curseswin=None):
    """ main function """

    if use_curses:
        global stdscr
        stdscr = curseswin
        curses.init_pair(1, curses.COLOR_BLUE, curses.COLOR_WHITE)
        curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
        curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
        curses.init_pair(4, curses.COLOR_RED, curses.COLOR_BLACK)
        curses.curs_set(0)
        #curses.resize_term(30,90)

    EventConnected.append(threading.Event())
    EventCompleted.append(threading.Event())
    EventCancel.append(threading.Event())

    with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
        executor.submit(ServerThread)
        executor.submit(ScreenThread)


def portnumber_type(portstr, min=1024, max=65535):
    val = int(portstr)
    if val >= min and val <= max:
        return val
    else:
        raise argparse.ArgumentTypeError('port not in range %s-%s'%(min,max))


if __name__ == '__main__':
    
    # check version
    if sys.version_info < (3,9):
        print("Error: Script expects python 3.9 or later.")
        sys.exit(0)

    # process command line parameters
    parser = argparse.ArgumentParser(description="TCP/IP server to display Image for Windows/Linux operation status.")
    parser.add_argument("-p", "--port", type=portnumber_type, default=54313, metavar="[1024-65535]", help="specify port for the connection (default: 54313)")
    parser.add_argument("-r", "--refresh", type=int, default=3, metavar="[1-60]", help="refresh rate in seconds for operation status (default is 3 if not specified or out of range)")
    args = parser.parse_args()
    global gPort
    gPort = int(args.port)
    global gRefreshRate
    gRefreshRate = int(args.refresh)
    if gRefreshRate < 1 or gRefreshRate > 60:
        gRefreshRate = 3

    if use_curses:
        curses.wrapper(main)
    else:
        main()